# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.core.api import RPCClient

from .. import exc
from ..interface import SearchInterface


class RemoteSearch(RPCClient):
    """Proxy to a remote search API"""

    backend_class = SearchInterface
    reraise_exceptions = [getattr(exc, exc_name) for exc_name in exc.__all__]



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
from textwrap import dedent
import types
import unittest

from elasticsearch.helpers.errors import BulkIndexError
import pytest

from swh.search.exc import SearchQuerySyntaxError
from swh.search.metrics import OPERATIONS_METRIC

from .test_search import CommonSearchTest

now = datetime.now(tz=timezone.utc).isoformat()
now_minus_5_days = (datetime.now(tz=timezone.utc) - timedelta(days=5)).isoformat()
now_plus_5_days = (datetime.now(tz=timezone.utc) + timedelta(days=5)).isoformat()

ORIGINS = [
    {
        "url": "http://foobar.1.com",
        "nb_visits": 1,
        "last_visit_date": now_minus_5_days,
        "last_eventful_visit_date": now_minus_5_days,
    },
    {
        "url": "http://foobar.2.com",
        "nb_visits": 2,
        "last_visit_date": now,
        "last_eventful_visit_date": now,
    },
    {
        "url": "http://foobar.3.com",
        "nb_visits": 3,
        "last_visit_date": now_plus_5_days,
        "last_eventful_visit_date": now_minus_5_days,
    },
    {
        "url": "http://barbaz.4.com",
        "nb_visits": 3,
        "last_visit_date": now_plus_5_days,
        "last_eventful_visit_date": now_minus_5_days,
    },
]


class CommonElasticsearchSearchTest(CommonSearchTest):
    """Tests shared between this module (direct ES backend test) and test_api_client.py
    (ES backend via HTTP test)"""

    def test_sort_by_and_limit_query(self):

        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_results(query, origin_indices):
            page = self.search.origin_search(url_pattern="foobar", query=query)
            results = [r["url"] for r in page.results]
            assert results == [ORIGINS[index]["url"] for index in origin_indices]

        _check_results("sort_by = [-visits]", [2, 1, 0])
        _check_results("sort_by = [last_visit]", [0, 1, 2])
        _check_results("sort_by = [-last_eventful_visit, visits]", [1, 0, 2])
        _check_results("sort_by = [last_eventful_visit,-last_visit]", [2, 0, 1])

        _check_results("sort_by = [-visits] limit = 1", [2])
        _check_results("sort_by = [last_visit] and limit = 2", [0, 1])
        _check_results("sort_by = [-last_eventful_visit, visits] limit = 3", [1, 0, 2])

    def test_search_ql_simple(self):
        self.search.origin_update(ORIGINS)
        self.search.flush()

        results = {
            r["url"]
            for r in self.search.origin_search(query='origin : "foobar"').results
        }
        assert results == {
            "http://foobar.1.com",
            "http://foobar.2.com",
            "http://foobar.3.com",
        }

    def test_search_ql_datetimes(self):
        self.search.origin_update(ORIGINS)
        self.search.flush()

        now_minus_5_minutes = (
            datetime.now(tz=timezone.utc) - timedelta(minutes=5)
        ).isoformat()
        now_plus_5_minutes = (
            datetime.now(tz=timezone.utc) + timedelta(minutes=5)
        ).isoformat()

        results = {
            r["url"]
            for r in self.search.origin_search(
                query=(
                    f"last_visit < {now_minus_5_minutes} "
                    f"or last_visit > {now_plus_5_minutes}"
                )
            ).results
        }
        assert results == {
            "http://foobar.1.com",
            "http://foobar.3.com",
            "http://barbaz.4.com",
        }

    def test_search_ql_dates(self):
        self.search.origin_update(ORIGINS)
        self.search.flush()

        now_minus_2_days = (
            (datetime.now(tz=timezone.utc) - timedelta(days=2)).date().isoformat()
        )
        now_plus_2_days = (
            (datetime.now(tz=timezone.utc) + timedelta(days=2)).date().isoformat()
        )

        results = {
            r["url"]
            for r in self.search.origin_search(
                query=(
                    f"last_visit < {now_minus_2_days} "
                    f"or last_visit > {now_plus_2_days}"
                )
            ).results
        }
        assert results == {
            "http://foobar.1.com",
            "http://foobar.3.com",
            "http://barbaz.4.com",
        }

    def test_search_ql_visited(self):
        self.search.origin_update(
            [
                {
                    "url": "http://foobar.1.com",
                    "has_visits": True,
                    "nb_visits": 1,
                    "last_visit_date": now_minus_5_days,
                    "last_eventful_visit_date": now_minus_5_days,
                },
                {
                    "url": "http://foobar.2.com",
                },
                {
                    "url": "http://foobar.3.com",
                    "has_visits": False,
                },
            ]
        )
        self.search.flush()

        assert {
            r["url"] for r in self.search.origin_search(query="visited = true").results
        } == {"http://foobar.1.com"}
        assert {
            r["url"] for r in self.search.origin_search(query="visited = false").results
        } == {"http://foobar.2.com", "http://foobar.3.com"}

        assert (
            self.search.origin_search(
                query="visited = true and visited = false"
            ).results
            == []
        )
        assert (
            self.search.origin_search(query="visited = false", with_visit=True).results
            == []
        )

    def test_query_syntax_error(self):
        self.search.origin_update(ORIGINS)
        self.search.flush()

        with pytest.raises(SearchQuerySyntaxError):
            self.search.origin_search(query="foobar")


class TestElasticsearchSearch(CommonElasticsearchSearchTest, unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _instantiate_search(self, swh_search, elasticsearch_host, mocker):
        self._elasticsearch_host = elasticsearch_host
        self.search = swh_search
        self.mocker = mocker

        # override self.search.origin_update to catch painless script errors
        # and pretty print them
        origin_update = self.search.origin_update

        def _origin_update(self, *args, **kwargs):
            script_error = False
            error_detail = ""
            try:
                origin_update(*args, **kwargs)
            except BulkIndexError as e:
                error = e.errors[0].get("update", {}).get("error", {}).get("caused_by")
                if error and "script_stack" in error:
                    script_error = True
                    error_detail = dedent(
                        f"""
                        Painless update script failed ({error.get('reason')}).
                        error type: {error.get('caused_by', {}).get('type')}
                        error reason: {error.get('caused_by', {}).get('reason')}
                        script stack:

                        """
                    )
                    error_detail += "\n".join(error["script_stack"])
                else:
                    raise e
            assert script_error is False, error_detail[1:]

        self.search.origin_update = types.MethodType(_origin_update, self.search)

    def reset(self):
        self.search.deinitialize()
        self.search.initialize()

    def test_metrics_update_duration(self):
        mock = self.mocker.patch("swh.search.metrics.statsd.timing")

        for url in ["http://foobar.bar", "http://foobar.baz"]:
            self.search.origin_update([{"url": url}])

        assert mock.call_count == 2

    def test_metrics_search_duration(self):
        mock = self.mocker.patch("swh.search.metrics.statsd.timing")

        for url_pattern in ["foobar", "foobaz"]:
            self.search.origin_search(url_pattern=url_pattern, with_visit=True)

        assert mock.call_count == 2

    def test_metrics_indexation_counters(self):
        mock_es = self.mocker.patch("elasticsearch.helpers.bulk")
        mock_es.return_value = 2, ["error"]

        mock_metrics = self.mocker.patch("swh.search.metrics.statsd.increment")

        self.search.origin_update([{"url": "http://foobar.baz"}])

        assert mock_metrics.call_count == 2

        mock_metrics.assert_any_call(
            OPERATIONS_METRIC,
            2,
            tags={
                "endpoint": "origin_update",
                "object_type": "document",
                "operation": "index",
            },
        )
        mock_metrics.assert_any_call(
            OPERATIONS_METRIC,
            1,
            tags={
                "endpoint": "origin_update",
                "object_type": "document",
                "operation": "index_error",
            },
        )

    def test_write_alias_usage(self):
        mock = self.mocker.patch("elasticsearch.helpers.bulk")
        mock.return_value = 2, ["result"]

        self.search.origin_update([{"url": "http://foobar.baz"}])

        assert mock.call_args[1]["index"] == "test-write"

    def test_read_alias_usage(self):
        mock = self.mocker.patch("elasticsearch.Elasticsearch.search")
        mock.return_value = {"hits": {"hits": []}}

        self.search.origin_search(url_pattern="foobar.baz")

        assert mock.call_args[1]["index"] == "test-read"



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
from datetime import datetime, timezone
import tempfile

from click.testing import CliRunner
from confluent_kafka import Producer
import pytest
from typing_extensions import Literal
import yaml

from swh.journal.serializers import value_to_kafka
from swh.model.hashutil import hash_to_bytes
from swh.search import get_search
from swh.search.cli import search_cli_group

CLI_CONFIG = """
search:
    cls: elasticsearch
    hosts:
    - '%(elasticsearch_host)s'
    indexes:
        origin:
            index: test
            read_alias: test-read
            write_alias: test-write
storage:
    cls: memory
"""

JOURNAL_OBJECTS_CONFIG_TEMPLATE = """
journal:
    brokers:
        - {broker}
    prefix: {prefix}
    group_id: {group_id}
"""


def invoke(catch_exceptions, args, config="", *, elasticsearch_host):
    runner = CliRunner()
    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        config_fd.write(
            (CLI_CONFIG + config) % {"elasticsearch_host": elasticsearch_host}
        )
        config_fd.seek(0)
        result = runner.invoke(search_cli_group, ["-C" + config_fd.name] + args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test__journal_client__origin(
    swh_search, elasticsearch_host: str, kafka_prefix: str, kafka_server
):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test search origin producer",
            "acks": "all",
        }
    )
    origin_foobar_baz = {
        "url": "http://foobar.baz",
    }
    value = value_to_kafka(origin_foobar_baz)
    topic = f"{kafka_prefix}.origin"
    producer.produce(topic=topic, key=b"bogus-origin", value=value)
    producer.flush()

    journal_objects_config = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker=kafka_server, prefix=kafka_prefix, group_id="test-consumer"
    )
    result = invoke(
        False,
        [
            "journal-client",
            "objects",
            "--stop-after-objects",
            "1",
            "--object-type",
            "origin",
            "--prefix",
            kafka_prefix,
        ],
        journal_objects_config,
        elasticsearch_host=elasticsearch_host,
    )

    # Check the output
    expected_output = "Processed 1 messages.\nDone.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    swh_search.flush()

    # searching origin without visit as requirement
    actual_page = swh_search.origin_search(url_pattern="foobar")
    # We find it
    assert actual_page.next_page_token is None
    assert actual_page.results == [origin_foobar_baz]

    # It's an origin with no visit, searching for it with visit
    actual_page = swh_search.origin_search(url_pattern="foobar", with_visit=True)
    # returns nothing
    assert actual_page.next_page_token is None
    assert actual_page.results == []


def test__journal_client__origin_visit_status(
    swh_search, elasticsearch_host, kafka_prefix: str, kafka_server
):
    """Subscribing to origin-visit-status should result in swh-search indexation"""
    origin_foobar = {"url": "http://baz.foobar"}
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test search origin visit status producer",
            "acks": "all",
        }
    )
    topic = f"{kafka_prefix}.origin_visit_status"
    value = value_to_kafka(
        {
            "origin": origin_foobar["url"],
            "visit": 1,
            "type": "git",
            "date": datetime.now(tz=timezone.utc),
            "snapshot": None,
            "status": "full",
        }
    )
    producer.produce(topic=topic, key=b"bogus-origin-visit-status", value=value)
    producer.flush()

    journal_objects_config = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker=kafka_server, prefix=kafka_prefix, group_id="test-consumer"
    )
    result = invoke(
        False,
        [
            "journal-client",
            "objects",
            "--stop-after-objects",
            "1",
            "--prefix",
            kafka_prefix,
            "--object-type",
            "origin_visit_status",
        ],
        journal_objects_config,
        elasticsearch_host=elasticsearch_host,
    )

    # Check the output
    expected_output = "Processed 1 messages.\nDone.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    swh_search.flush()

    # Both search returns the visit
    actual_page = swh_search.origin_search(url_pattern="foobar", with_visit=False)
    assert actual_page.next_page_token is None
    assert actual_page.results == [origin_foobar]

    actual_page = swh_search.origin_search(url_pattern="foobar", with_visit=True)
    assert actual_page.next_page_token is None
    assert actual_page.results == [origin_foobar]


@pytest.mark.parametrize("metadata_source", ["intrinsic", "extrinsic"])
def test__journal_client__origin_metadata(
    swh_search,
    elasticsearch_host,
    kafka_prefix: str,
    kafka_server,
    metadata_source: Literal["intrinsic", "extrinsic"],
):
    """Subscribing to origin-intrinsic-metadata should result in swh-search indexation"""
    origin_foobar = {"url": "https://github.com/clojure/clojure"}

    origin_metadata = {
        "id": origin_foobar["url"],
        "metadata": {
            "name": "clojure",
            "type": "SoftwareSourceCode",
            "license": "http://opensource.org/licenses/eclipse-1.0.php",
            "version": "1.10.2-master-SNAPSHOT",
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "identifier": "org.clojure",
            "description": "Clojure core environment and runtime library.",
            "codeRepository": "https://repo.maven.apache.org/maven2/org/clojure/clojure",  # noqa
        },
        "indexer_configuration_id": 1,
    }
    if metadata_source == "intrinsic":
        origin_metadata.update(
            {
                "from_revision": hash_to_bytes(
                    "f47c139e20970ee0852166f48ee2a4626632b86e"
                ),
                "mappings": ["maven"],
            }
        )
    elif metadata_source == "extrinsic":
        origin_metadata.update(
            {
                "from_revision": hash_to_bytes(
                    "f47c139e20970ee0852166f48ee2a4626632b86e"
                ),
                "mappings": ["github"],
            }
        )
    else:
        assert False, metadata_source

    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test search origin intrinsic metadata producer",
            "acks": "all",
        }
    )
    topic = f"{kafka_prefix}.origin_{metadata_source}_metadata"
    value = value_to_kafka(origin_metadata)
    producer.produce(topic=topic, key=b"bogus-origin-metadata", value=value)
    producer.flush()

    journal_objects_config = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker=kafka_server, prefix=kafka_prefix, group_id="test-consumer"
    )
    result = invoke(
        False,
        [
            "journal-client",
            "objects",
            "--stop-after-objects",
            "1",
            "--object-type",
            f"origin_{metadata_source}_metadata",
        ],
        journal_objects_config,
        elasticsearch_host=elasticsearch_host,
    )

    # Check the output
    expected_output = "Processed 1 messages.\nDone.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    swh_search.flush()

    # search without visit returns the metadata
    actual_page = swh_search.origin_search(url_pattern="clojure", with_visit=False)
    assert actual_page.next_page_token is None
    assert actual_page.results == [origin_foobar]

    # no visit associated so it does not return anything
    actual_page = swh_search.origin_search(url_pattern="clojure", with_visit=True)
    assert actual_page.next_page_token is None
    assert actual_page.results == []


def test__journal_client__missing_main_journal_config_key(elasticsearch_host):
    """Missing configuration on journal should raise"""
    with pytest.raises(KeyError, match="journal"):
        invoke(
            catch_exceptions=False,
            args=[
                "journal-client",
                "objects",
                "--stop-after-objects",
                "1",
            ],
            config="",  # missing config will make it raise
            elasticsearch_host=elasticsearch_host,
        )


def test__journal_client__missing_journal_config_keys(elasticsearch_host):
    """Missing configuration on mandatory journal keys should raise"""
    kafka_prefix = "swh.journal.objects"
    journal_objects_config = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker="192.0.2.1", prefix=kafka_prefix, group_id="test-consumer"
    )
    journal_config = yaml.safe_load(journal_objects_config)

    for key in journal_config["journal"].keys():
        if key == "prefix":  # optional
            continue
        cfg = copy.deepcopy(journal_config)
        del cfg["journal"][key]  # make config incomplete
        yaml_cfg = yaml.dump(cfg)

        with pytest.raises(TypeError, match=f"{key}"):
            invoke(
                catch_exceptions=False,
                args=[
                    "journal-client",
                    "objects",
                    "--stop-after-objects",
                    "1",
                    "--prefix",
                    kafka_prefix,
                    "--object-type",
                    "origin_visit_status",
                ],
                config=yaml_cfg,  # incomplete config will make the cli raise
                elasticsearch_host=elasticsearch_host,
            )


def test__journal_client__missing_prefix_config_key(
    swh_search, elasticsearch_host, kafka_server
):
    """Missing configuration on mandatory prefix key should raise"""

    journal_cfg_template = """
journal:
    brokers:
        - {broker}
    group_id: {group_id}
    """

    journal_cfg = journal_cfg_template.format(
        broker=kafka_server, group_id="test-consumer"
    )

    with pytest.raises(ValueError, match="prefix"):
        invoke(
            False,
            # Missing --prefix (and no config key) will make the cli raise
            [
                "journal-client",
                "objects",
                "--stop-after-objects",
                "1",
                "--object-type",
                "origin_visit_status",
            ],
            journal_cfg,
            elasticsearch_host=elasticsearch_host,
        )


def test__journal_client__missing_object_types_config_key(
    swh_search, elasticsearch_host, kafka_server
):
    """Missing configuration on mandatory object-types key should raise"""

    journal_cfg_template = """
journal:
    brokers:
        - {broker}
    prefix: swh.journal.objects
    group_id: {group_id}
    """

    journal_cfg = journal_cfg_template.format(
        broker=kafka_server, group_id="test-consumer"
    )

    with pytest.raises(ValueError, match="object_types"):
        invoke(
            False,
            # Missing --object-types (and no config key) will make the cli raise
            ["journal-client", "objects", "--stop-after-objects", "1"],
            journal_cfg,
            elasticsearch_host=elasticsearch_host,
        )


def test__initialize__with_index_name(elasticsearch_host):
    """Initializing the index with an index name should create the right index"""

    search = get_search(
        "elasticsearch",
        hosts=[elasticsearch_host],
        indexes={"origin": {"index": "test"}},
    )

    assert search._get_origin_index() == "test"
    assert search._get_origin_read_alias() == "origin-read"
    assert search._get_origin_write_alias() == "origin-write"


def test__initialize__with_read_alias(elasticsearch_host):
    """Initializing the index with a search alias name should create
    the right search alias"""

    search = get_search(
        "elasticsearch",
        hosts=[elasticsearch_host],
        indexes={"origin": {"read_alias": "test"}},
    )

    assert search._get_origin_index() == "origin"
    assert search._get_origin_read_alias() == "test"
    assert search._get_origin_write_alias() == "origin-write"


def test__initialize__with_write_alias(elasticsearch_host):
    """Initializing the index with an indexing alias name should create
    the right indexing alias"""

    search = get_search(
        "elasticsearch",
        hosts=[elasticsearch_host],
        indexes={"origin": {"write_alias": "test"}},
    )

    assert search._get_origin_index() == "origin"
    assert search._get_origin_read_alias() == "origin-read"
    assert search._get_origin_write_alias() == "test"



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
from typing import Any, Dict

from swh.core import config
from swh.core.api import RPCServerApp
from swh.core.api import encode_data_server as encode_data
from swh.core.api import error_handler
from swh.search.metrics import timed

from .. import get_search
from ..exc import SearchException
from ..interface import SearchInterface

logger = logging.getLogger(__name__)


def _get_search():
    global search
    if not search:
        search = get_search(**app.config["search"])

    return search


app = RPCServerApp(__name__, backend_class=SearchInterface, backend_factory=_get_search)

search = None


@app.errorhandler(SearchException)
def search_error_handler(exception):
    return error_handler(exception, encode_data, status_code=400)


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


@app.route("/")
@timed
def index():
    return "SWH Search API server"


@app.before_first_request
def initialized_index():
    search = _get_search()
    if app.config["search"]["cls"] == "elasticsearch":
        logger.info("Initializing indexes with configuration: ", search.origin_config)
    search.initialize()


api_cfg = None


def load_and_check_config(config_file: str) -> Dict[str, Any]:
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_file: Path to the configuration file to load
        type: configuration type. For 'local' type, more
                    checks are done.

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_file:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_file):
        raise FileNotFoundError("Configuration file %s does not exist" % (config_file,))

    cfg = config.read(config_file)
    if "search" not in cfg:
        raise KeyError("Missing 'search' configuration")

    return cfg


def make_app_from_configfile():
    """Run the WSGI app from the webserver, loading the configuration from
    a configuration file.

    SWH_CONFIG_FILENAME environment variable defines the
    configuration path to load.

    """
    global api_cfg
    if not api_cfg:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")
        api_cfg = load_and_check_config(config_file)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app
